package formula

case class Formula(symbol: Char, children: Vector[Formula]) extends Ordered[Formula] {
  def isAtom = children.isEmpty
  def isConnective = children.nonEmpty
  override def toString = symbol + childrenString
  def childrenString = if (isAtom) "" else children.mkString("(",",",")")
  def sort: Formula = if (isAtom) this else Formula(symbol,children.map(_.sort).sorted)
  override def compare(that: Formula): Int = {
    val symbolCompare = symbol.compare(that.symbol)
    if (symbolCompare != 0) return symbolCompare
    val numChildrenCompare = children.size.compare(that.children.size)
    if (numChildrenCompare != 0) return numChildrenCompare
    for (i <- 0 until children.size) {
      val childCompare = children(i).compare(that.children(i))
      if (childCompare != 0) return childCompare
    }
    0
  }
  def compress: Formula = {  // remove adjacent same-connectives, e.g. @(@(A,B),+(C,D)) -> @(A,B,+(C,D))
    if (isAtom) return this
    val newChildren: Vector[Formula] = children.flatMap { case rawChild =>
        val child = rawChild.compress
        if (child.symbol == symbol) child.children
        else Vector(child)
    }
    Formula(symbol,newChildren)
  }
}

object PosAtom extends Formula(Formula.positiveAtomSymbol,Vector.empty)
object NegAtom extends Formula(Formula.negativeAtomSymbol,Vector.empty)

object Atom { def apply(symbol: Char): Formula = if (symbol == Formula.positiveAtomSymbol) PosAtom else NegAtom }

object Tensor { def apply(children: Vector[Formula]): Formula = Formula(Formula.tensorSymbol,children) }
object Par    { def apply(children: Vector[Formula]): Formula = Formula(Formula.parSymbol   ,children) }
object With   { def apply(children: Vector[Formula]): Formula = Formula(Formula.withSymbol  ,children) }
object Plus   { def apply(children: Vector[Formula]): Formula = Formula(Formula.plusSymbol  ,children) }

object Formula {

  val positiveAtomSymbol = '1'
  val negativeAtomSymbol = '0'
  val tensorSymbol = '@'
  val parSymbol = '#'
  val withSymbol = '&'
  val plusSymbol = '+'

  val isAtomSymbol = Set(positiveAtomSymbol,negativeAtomSymbol)
  val isConnectiveSymbol = Set(tensorSymbol,parSymbol,withSymbol,plusSymbol)

  case class ParsedFormulaFromHead(formula: Formula, remainder: String)

  case class ParsedChildrenFromHead(children: Vector[Formula], remainder: String)

  def parseFromFront(s: String): ParsedFormulaFromHead = {
    val symbol: Char = s(0)
    if (isAtomSymbol(symbol)) ParsedFormulaFromHead(Atom(symbol),s.tail)  // tail either starts with comma ',' or is empty
    else {
      //      println("    isConnectiveSymbol "+symbol)
      val parsedChildren = parseChildrenInParensFromHead(s.tail) // tail starts "(", remainder either starts comma ',' or is empty
      ParsedFormulaFromHead(Formula(symbol,parsedChildren.children),parsedChildren.remainder)
    }
  }

  def parseChildrenInParensFromHead(s: String): ParsedChildrenFromHead = {
    var remainder = s
    var children = Vector[Formula]()
    while (remainder.head != ')') {  // starts with comma or open '('
    val parsedChild = parseFromFront(remainder.tail)  // tail => drop leading '(' or comma
      remainder = parsedChild.remainder
      children :+= parsedChild.formula
    }
    ParsedChildrenFromHead(children,remainder.tail) // _.tail => drop the ')'   hence remainder.tail starts with comma or is empty
  }

  def parse(s: String): Formula = parseFromFront(s).formula

  def testParse: Unit = {
    val cases = List(
      (""+positiveAtomSymbol,PosAtom)
      ,
      (""+negativeAtomSymbol,NegAtom)
      ,
      (tensorSymbol+"("+negativeAtomSymbol+","+positiveAtomSymbol+")",Tensor(Vector(NegAtom,PosAtom)))
      ,
      (tensorSymbol+"("+negativeAtomSymbol+","+positiveAtomSymbol+","+negativeAtomSymbol+")",Tensor(Vector(NegAtom,PosAtom,NegAtom)))
      ,
      (withSymbol+"("+tensorSymbol+"("+negativeAtomSymbol+","+positiveAtomSymbol+","+negativeAtomSymbol+"),"+plusSymbol+"("+positiveAtomSymbol+","+negativeAtomSymbol+"))",
        With(Vector(Tensor(Vector(NegAtom,PosAtom,NegAtom)),Plus(Vector(PosAtom,NegAtom)))))
    )
    for ((input,expect) <- cases) {
      println()
      println("input  "+input)
      println("expect "+expect)
      val got = parse(input)
      println("got    "+got)
      require(expect == got)
    }
  }

  def testCompare: Unit = {
    val cases = List(
      "1 0"
      ,
      "@(1,1) @(0,0)"
      ,
      "@(0,1) @(1,0)"
      ,
      "@(1,1) #(1,1)"
      ,
      "@(1,1) @(1,1,1)"
    )
    for (c <- cases) {
      println()
      val pair = c.split("\\s+")
      val f = parse(pair(0))
      val g = parse(pair(1))
      println("f = "+f)
      println("g = "+g)
      val s = f.sort
      val t = g.sort
      println("s = f.sort = "+s)
      println("t = g.sort = "+t)
      println("s < t  =  "+(s<t))
      println("s.compare(t) = "+s.compare(t))
    }

  }

  def testCompress: Unit = {
    val cases = List("@(@(0,1),+(1,0),@(@(0,0,0),@(1,1)))")
    for (c <- cases) {
      val f = parse(c)
      val cf = f.compress
      println(f + " -> " + cf)
    }
  }

  def main(args: Array[String]) {
//    testParse
//    testCompare
    testCompress
  }


}